import { basename, extname, resolve, dirname, join } from 'path';
import { createPage } from '@vuepress/core';
import _docgen, { extractConfig } from 'vue-docgen-cli';
import WebpackConfig from 'webpack-chain';
import glob from 'globby';
import { defu } from 'defu';
import chokidar from 'chokidar';
import _originalComponentTemplate from 'vue-docgen-cli/lib/templates/component.js';
import 'fs/promises';
import 'fs';
import matter from 'gray-matter';

const webpackHandleResolve = async ({
  app,
  config,
  isServer
}) => {
  config.resolve.alias.set("@source", app.dir.source()).set("@temp", app.dir.temp()).set("@internal", app.dir.temp("internal"));
  config.merge({
    resolve: {
      extensionAlias: {
        ".js": [".js", ".ts"],
        ".mjs": [".mjs", ".mts"]
      },
      extensions: [".js", ".jsx", ".ts", ".tsx", ".vue", ".json"]
    }
  });
  config.resolve.extensions.merge([
    ".js",
    ".jsx",
    ".ts",
    ".tsx",
    ".vue",
    ".json"
  ]);
  const aliasResult = await app.pluginApi.hooks.alias.process(app, isServer);
  aliasResult.forEach(
    (aliasObject) => Object.entries(aliasObject).forEach(([key, value]) => {
      config.resolve.alias.set(key, value);
    })
  );
};

const sleep = (ms = 0) => new Promise((res) => setTimeout(res, ms));

const editLinkRegexp = /^<a href="[^"]+" class="docgen-edit-link">[^<]+<\/a>/im;
const extractAndCutFrontmatter = (doc, grayMatterOptions, content = "") => {
  const {
    data: topContentFrontmatter = {},
    content: cuttedTopContent
    // Without .trim() matter ignore some files
  } = matter(content.trim(), grayMatterOptions);
  let frontmatter = topContentFrontmatter;
  doc.docsBlocks?.forEach((blockContent, index) => {
    let formattedContent = blockContent.trim();
    const editLinkMatch = formattedContent.match(editLinkRegexp);
    if (editLinkMatch) {
      formattedContent = formattedContent.replace(editLinkRegexp, "").trim();
    }
    let {
      data,
      content: cuttedBlockContent
    } = matter(formattedContent, grayMatterOptions);
    if (editLinkMatch) {
      cuttedBlockContent = [
        editLinkMatch[0],
        cuttedBlockContent
      ].join("\n\n");
    }
    frontmatter = defu(data, frontmatter);
    doc.docsBlocks[index] = cuttedBlockContent;
  });
  return {
    content: matter.stringify(cuttedTopContent, frontmatter, grayMatterOptions),
    frontmatter
  };
};

const defaultGetDestFile = (file, config) => {
  if (basename(file, extname(file)).toLowerCase() === "index")
    return resolve(config.outDir, dirname(file), "README.md");
  return resolve(config.outDir, file).replace(/\.\w+$/, ".md");
};

const reResolveAppPages = async (app) => {
  const pageFilePaths = await glob(app.options.pagePatterns, {
    absolute: true,
    cwd: app.dir.source()
  });
  const createdPages = await Promise.all(
    pageFilePaths.filter(
      (filePath) => (
        // Ignore already added pages
        !app.pages.some((page) => page.filePath === filePath)
      )
    ).map((filePath) => createPage(app, { filePath }))
  );
  app.pages.push(...createdPages);
};

const originalComponentTemplate = _originalComponentTemplate.default;
const templateComponent = (grayMatterOptions) => (renderedUsage, doc, config, fileName, requiresMd, subTemplateOptions) => {
  if (subTemplateOptions.isSubComponent) {
    return originalComponentTemplate(
      renderedUsage,
      doc,
      config,
      fileName,
      requiresMd,
      subTemplateOptions
    );
  }
  const originalResultWithoutDocsBlocks = originalComponentTemplate(
    renderedUsage,
    {
      ...doc,
      docsBlocks: []
    },
    config,
    fileName,
    requiresMd,
    subTemplateOptions
  );
  let {
    content
  } = extractAndCutFrontmatter(
    doc,
    grayMatterOptions,
    originalResultWithoutDocsBlocks
  );
  if (doc.docsBlocks)
    content += "---\n" + doc.docsBlocks.join("\n---\n");
  return content;
};

const tmpFolderName = "docgen";

const docgen = _docgen.default;
const VueDocgenPlugin = ({
  docgenCliConfig = {},
  docgenCliConfigPath,
  groups = [{ components: ["**/components/**/*.vue", "!**/node_modules/**", "!**/.vuepress/**"] }],
  stateless = true
}) => {
  if (!Array.isArray(groups))
    groups = [groups];
  const normalizedGroups = groups.map((stringOrObject) => {
    if (typeof stringOrObject === "string")
      return {
        components: stringOrObject
      };
    return stringOrObject;
  });
  return {
    name: "vuepress-plugin-docgen",
    onInitialized: async (app) => {
      const { grayMatterOptions } = app.options.markdown.frontmatter || {};
      const tmpFolder = join(app.options.temp, tmpFolderName);
      const rootFolder = app.dir.source();
      const webpackConfig = new WebpackConfig();
      await webpackHandleResolve({ app, config: webpackConfig, isServer: true });
      const baseDocgenCliConfig = defu(docgenCliConfig, {
        apiOptions: {
          ...webpackConfig.toConfig().resolve
        },
        templates: {
          component: templateComponent(grayMatterOptions)
        },
        getDestFile: defaultGetDestFile
      }, await extractConfig(process.cwd(), app.env.isDev, docgenCliConfigPath, []));
      await Promise.all(normalizedGroups.map(async ({
        root: componentsRoot,
        components,
        outDir: rawOutDir = "",
        docgenCliConfig: groupDocgenCliConfig = {}
      }) => {
        const outDir = stateless ? resolve(tmpFolder, rawOutDir) : resolve(rootFolder, rawOutDir);
        const config = defu({
          componentsRoot,
          components,
          outDir
        }, baseDocgenCliConfig, groupDocgenCliConfig);
        await docgen(config);
      }));
      await sleep(100);
      if (!stateless) {
        await reResolveAppPages(app);
        return;
      }
      const docFiles = await glob("**/*.md", {
        cwd: tmpFolder
      });
      await Promise.all(docFiles.map(async (relativeDocPath) => {
        let normalizedRelativeDocPath = relativeDocPath;
        const docBasename = basename(relativeDocPath);
        normalizedRelativeDocPath = normalizedRelativeDocPath.slice(0, -3);
        if (docBasename.toLowerCase() === "readme.md")
          normalizedRelativeDocPath = normalizedRelativeDocPath.replace(/readme$/i, "");
        if (normalizedRelativeDocPath.endsWith("/"))
          normalizedRelativeDocPath = normalizedRelativeDocPath.slice(0, -1);
        const page = await createPage(app, {
          path: "/" + normalizedRelativeDocPath,
          filePath: join(tmpFolder, relativeDocPath),
          isDocgenPage: true,
          frontmatter: {
            // Disable editLink by default for stateless
            // because it created in tmp folder
            editLink: false
          }
        });
        if (page.permalink)
          page.path = page.permalink;
        app.pages.push(page);
      }));
    },
    onWatched(app, watchers, restart) {
      if (!stateless)
        return;
      const tmpFolder = join(app.options.temp, tmpFolderName);
      const watcher = chokidar.watch("**/*.md", {
        cwd: tmpFolder,
        ignoreInitial: true
      });
      watcher.on("change", () => restart());
      watchers.push(watcher);
    }
  };
};

export { VueDocgenPlugin, extractAndCutFrontmatter, templateComponent };
